﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;

namespace Praktikum1
{
    // Die Kommentare hier sind größtenteils nur didaktisch.
    // Die meisten würden nicht in einem echten Programm stehen.
    // Und natürlich würde man englische Bezeichnungen verwenden.

    public partial class MainWindow : Window
    {
        const int zahlPersonen = 100000;
        const int zahlAnfangsInfizierte = 100;
        const int zahlDerTreffenProPersonUndTag = 10;
        const double infektionsWahrscheinlichkeitProTreffen = 0.02;

        DateTime simulationszeitpunkt;
        int simulationsschritt;

        Person[] personen = new Person[zahlPersonen];

        // benötigt ein "using", z.B. per Glühlampen-Popup
        DispatcherTimer timer = new DispatcherTimer();

        static Random würfel = new Random();

        public MainWindow()
        {
            InitializeComponent();

            timer.Interval = TimeSpan.FromMilliseconds(100);
            timer.Tick += Animiere; // der Name der Methode, also keine runden Klammern!
        }

        private void buttonStart_Click(object sender, RoutedEventArgs e)
        {
            simulationszeitpunkt = DateTime.Now;
            simulationsschritt = 0;
            kurveGenesene.Points.Clear();
            kurveInfizierte.Points.Clear();

            for (int i = 0; i < zahlPersonen; i++)
            {
                personen[i] = new Person();
            }

            for (int i = 0; i < zahlAnfangsInfizierte; i++)
            {
                personen[i].ErhalteErreger(simulationszeitpunkt);
            }

            timer.Start();
            buttonStart.IsEnabled = false;
            buttonStop.IsEnabled = true;
        }

        private void buttonStop_Click(object sender, RoutedEventArgs e)
        {
            timer.Stop();
            buttonStart.IsEnabled = true;
            buttonStop.IsEnabled = false;
        }

        void Animiere(object sender, EventArgs e)
        {
            // Immer exakt die vorgebene Zahl an Treffen pro Person zu haben,
            // ist aufwändig und für die Simulation nicht nötig.
            // Eine nicht-exakte Lösung ist für dieses Praktikum ok.
            // Aber hier ist eine exakte Lösung:

            // Die folgenden Arrays werden so gepflegt, dass am Anfang lückenlos
            // die Personen stehen, die noch nicht genügend Treffen haben.
            // Die Arrays gelten nur bis zum Index zahlDerNochTreffbarenPersonen - 1.
            // Dahinter stehen Überbleibsel.
            Person[] nochTreffbarePersonen = new Person[zahlPersonen];
            int[] zahlDerVerbleibendenTreffen = new int[zahlPersonen];
            int zahlDerNochTreffbarenPersonen = zahlPersonen;

            for (int i = 0; i < zahlPersonen; i++)
            {
                nochTreffbarePersonen[i] = personen[i];
                zahlDerVerbleibendenTreffen[i] = zahlDerTreffenProPersonUndTag;
            }

            while(zahlDerNochTreffbarenPersonen > 0)
            {
                // Die letzte der möglichen Personen
                // trifft sich mit einer zufälligen Person.
                int i = zahlDerNochTreffbarenPersonen - 1;
                int j = würfel.Next(zahlDerNochTreffbarenPersonen - 1);

                zahlDerVerbleibendenTreffen[i]--;
                if (zahlDerVerbleibendenTreffen[i] == 0)
                {
                    // Dies verschiebt auch das effektive Ende der Arrays!
                    zahlDerNochTreffbarenPersonen--;
                }

                zahlDerVerbleibendenTreffen[j]--;
                if (zahlDerVerbleibendenTreffen[j] == 0)
                {
                    zahlDerNochTreffbarenPersonen--;

                    // Um die Person in j zu entfernen,
                    // überschreibe sie mit den Daten
                    // der hintersten verbleibenden Person.
                    nochTreffbarePersonen[j] = nochTreffbarePersonen[zahlDerNochTreffbarenPersonen];
                    zahlDerVerbleibendenTreffen[j] = zahlDerVerbleibendenTreffen[zahlDerNochTreffbarenPersonen];
                }

                if (würfel.NextDouble() < infektionsWahrscheinlichkeitProTreffen)
                {
                    Person pi = nochTreffbarePersonen[i];
                    Person pj = nochTreffbarePersonen[j];

                    if (pi.KannInfizieren(simulationszeitpunkt))
                    {
                        pj.ErhalteErreger(simulationszeitpunkt);
                    }
                    else if (pj.KannInfizieren(simulationszeitpunkt))
                    {
                        pi.ErhalteErreger(simulationszeitpunkt);
                    }
                }
            }

            // Jetzt am Anfang des Semesters
            // dies natürlich mit einer for-Schleife:
            int zahlInfizierte = personen.Count(p => p.IstInfiziert(simulationszeitpunkt));
            int zahlGenesene = personen.Count(p => p.IstGenesen(simulationszeitpunkt));

            // Die y-Achse der Zeichenfunktionen zeigt nach unten!
            kurveInfizierte.Points.Add(new Point(5 * simulationsschritt, zeichenfläche.ActualHeight * (1.0 - zahlInfizierte / (double) zahlPersonen)));
            kurveGenesene.Points.Add(new Point(5 * simulationsschritt, zeichenfläche.ActualHeight * (1.0 - zahlGenesene / (double) zahlPersonen)));

            labelEnde.Content = simulationszeitpunkt.ToString("d");

            simulationszeitpunkt += TimeSpan.FromDays(1);
            simulationsschritt++;
        }
    }
}